@@ -8,14 +8,16 @@ import fnmatch
 import inspect
 import os
 import pickle
+from types import MethodType
+from typing import (Any, Callable, Dict, Generator, Iterable, List, Optional, Set, Type, TYPE_CHECKING, Union)
 
 import yaml
 
 from . import loaders
 from . import futures
 from . import utils
-from . import base
-from .base import super_check
+from .utils import PID_TYPE, SAVED_STATE_TYPE
+from .base.utils import super_check, call_with_super_check
 
 __all__ = [
     'Bundle', 'Persister', 'PicklePersister', 'auto_persist', 'Savable', 'SavableFuture', 'LoadSaveContext',
@@ -24,30 +26,36 @@ __all__ = [
 
 PersistedCheckpoint = collections.namedtuple('PersistedCheckpoint', ['pid', 'tag'])
 
+if TYPE_CHECKING:
+    from .processes import Process  # pylint: disable=cyclic-import
+
 
 class Bundle(dict):
 
-    def __init__(self, savable, save_context=None):
+    def __init__(self, savable: 'Savable', save_context: Optional['LoadSaveContext'] = None, dereference: bool = False):
         """
         Create a bundle from a savable.  Optionally keep information about the
         class loader that can be used to load the classes in the bundle.
 
         :param savable: The savable object to bundle
-        :type savable: :class:`Savable`
         :param save_context: The optional save context to use
-        :type save_context: :class:`LoadSaveContext`
+        :param dereference: Remove refrences from the data, by deep copying
+
         """
         super().__init__()
-        self.update(savable.save(save_context))
+        if dereference:
+            self.update(copy.deepcopy(savable.save(save_context)))
+        else:
+            self.update(savable.save(save_context))
 
-    def unbundle(self, load_context=None):
+    def unbundle(self, load_context: Optional['LoadSaveContext'] = None) -> 'Savable':
         """
         This method loads the class of the object and calls its recreate_from
         method passing the positional and keyword arguments.
 
         :param load_context: The optional load context
         :return: An instance of the Savable
-        :rtype: :class:`Savable`
+
         """
         return Savable.load(self, load_context)
 
@@ -55,11 +63,11 @@ class Bundle(dict):
 _BUNDLE_TAG = '!plumpy:Bundle'
 
 
-def _bundle_representer(dumper, node):
+def _bundle_representer(dumper: yaml.Dumper, node: Any) -> Any:
     return dumper.represent_mapping(_BUNDLE_TAG, node)
 
 
-def _bundle_constructor(loader, data):
+def _bundle_constructor(loader: yaml.Loader, data: Any) -> Generator[Bundle, None, None]:
     result = Bundle.__new__(Bundle)
     yield result
     mapping = loader.construct_mapping(data)
@@ -73,7 +81,7 @@ yaml.add_constructor(_BUNDLE_TAG, _bundle_constructor)
 class Persister(metaclass=abc.ABCMeta):
 
     @abc.abstractmethod
-    def save_checkpoint(self, process, tag=None):
+    def save_checkpoint(self, process: 'Process', tag: Optional[str] = None) -> None:
         """
         Persist a Process instance
 
@@ -84,7 +92,7 @@ class Persister(metaclass=abc.ABCMeta):
         """
 
     @abc.abstractmethod
-    def load_checkpoint(self, pid, tag=None):
+    def load_checkpoint(self, pid: PID_TYPE, tag: Optional[str] = None) -> Bundle:
         """
         Load a process from a persisted checkpoint by its process id
 
@@ -92,21 +100,21 @@ class Persister(metaclass=abc.ABCMeta):
         :param tag: optional checkpoint identifier to allow retrieving
             a specific sub checkpoint for the corresponding process
         :return: a bundle with the process state
-        :rtype: :class:`plumpy.Bundle`
+
         :raises: :class:`plumpy.PersistenceError` Raised if there was a problem loading the checkpoint
         """
 
     @abc.abstractmethod
-    def get_checkpoints(self):
+    def get_checkpoints(self) -> List[PersistedCheckpoint]:
         """
         Return a list of all the current persisted process checkpoints
         with each element containing the process id and optional checkpoint tag
 
-        :return: list of PersistedCheckpoint tuples
+        :return: list of PersistedCheckpoint
         """
 
     @abc.abstractmethod
-    def get_process_checkpoints(self, pid):
+    def get_process_checkpoints(self, pid: PID_TYPE) -> List[PersistedCheckpoint]:
         """
         Return a list of all the current persisted process checkpoints for the
         specified process with each element containing the process id and
@@ -117,7 +125,7 @@ class Persister(metaclass=abc.ABCMeta):
         """
 
     @abc.abstractmethod
-    def delete_checkpoint(self, pid, tag=None):
+    def delete_checkpoint(self, pid: PID_TYPE, tag: Optional[str] = None) -> None:
         """
         Delete a persisted process checkpoint. No error will be raised if
         the checkpoint does not exist
@@ -128,7 +136,7 @@ class Persister(metaclass=abc.ABCMeta):
         """
 
     @abc.abstractmethod
-    def delete_process_checkpoints(self, pid):
+    def delete_process_checkpoints(self, pid: PID_TYPE) -> None:
         """
         Delete all persisted checkpoints related to the given process id
 
@@ -146,7 +154,7 @@ class PicklePersister(Persister):
     in pickles on a filesystem.
     """
 
-    def __init__(self, pickle_directory):
+    def __init__(self, pickle_directory: str):
         """
         Instantiate a PicklePersister object that will persist processes by
         writing their bundles to a pickle in a directory specified by the
@@ -164,7 +172,7 @@ class PicklePersister(Persister):
         self._pickle_directory = pickle_directory
 
     @staticmethod
-    def ensure_pickle_directory(dirpath):
+    def ensure_pickle_directory(dirpath: str) -> None:
         """
         Will attempt to create the directory at dirpath and raise if it fails, except
         if the exception arose because the directory already existed
@@ -176,13 +184,13 @@ class PicklePersister(Persister):
                 raise
 
     @staticmethod
-    def load_pickle(filepath):
+    def load_pickle(filepath: str) -> 'PersistedPickle':
         """
         Load a pickle from disk
 
         :param filepath: absolute filepath to the pickle
         :returns: the loaded pickle
-        :rtype: PersistedPickle
+
         """
         with open(filepath, 'r+b') as handle:
             persisted_pickle = pickle.load(handle)
@@ -190,7 +198,7 @@ class PicklePersister(Persister):
         return persisted_pickle
 
     @staticmethod
-    def pickle_filename(pid, tag=None):
+    def pickle_filename(pid: PID_TYPE, tag: Optional[str] = None) -> str:
         """
         Returns the relative filepath of the pickle for the given process id
         and optional checkpoint tag
@@ -202,14 +210,14 @@ class PicklePersister(Persister):
 
         return filename
 
-    def _pickle_filepath(self, pid, tag=None):
+    def _pickle_filepath(self, pid: PID_TYPE, tag: Optional[str] = None) -> str:
         """
         Returns the full filepath of the pickle for the given process id
         and optional checkpoint tag
         """
         return os.path.join(self._pickle_directory, PicklePersister.pickle_filename(pid, tag))
 
-    def save_checkpoint(self, process, tag=None):
+    def save_checkpoint(self, process: 'Process', tag: Optional[str] = None) -> None:
         """
         Persist a process to a pickle on disk
 
@@ -224,7 +232,7 @@ class PicklePersister(Persister):
         with open(self._pickle_filepath(process.pid, tag), 'w+b') as handle:
             pickle.dump(persisted_pickle, handle)
 
-    def load_checkpoint(self, pid, tag=None):
+    def load_checkpoint(self, pid: PID_TYPE, tag: Optional[str] = None) -> Bundle:
         """
         Load a process from a persisted checkpoint by its process id
 
@@ -232,19 +240,19 @@ class PicklePersister(Persister):
         :param tag: optional checkpoint identifier to allow retrieving
             a specific sub checkpoint for the corresponding process
         :return: a bundle with the process state
-        :rtype: :class:`plumpy.Bundle`
+
         """
         filepath = self._pickle_filepath(pid, tag)
         checkpoint = PicklePersister.load_pickle(filepath)
 
         return checkpoint.bundle
 
-    def get_checkpoints(self):
+    def get_checkpoints(self) -> List[PersistedCheckpoint]:
         """
         Return a list of all the current persisted process checkpoints
         with each element containing the process id and optional checkpoint tag
 
-        :return: list of PersistedCheckpoint tuples
+        :return: list of PersistedCheckpoint
         """
         checkpoints = []
         file_pattern = '*.{}'.format(_PICKLE_SUFFIX)
@@ -257,18 +265,18 @@ class PicklePersister(Persister):
 
         return checkpoints
 
-    def get_process_checkpoints(self, pid):
+    def get_process_checkpoints(self, pid: PID_TYPE) -> List[PersistedCheckpoint]:
         """
         Return a list of all the current persisted process checkpoints for the
         specified process with each element containing the process id and
         optional checkpoint tag
 
         :param pid: the process pid
-        :return: list of PersistedCheckpoint tuples
+        :return: list of PersistedCheckpoint
         """
         return [c for c in self.get_checkpoints() if c.pid == pid]
 
-    def delete_checkpoint(self, pid, tag=None):
+    def delete_checkpoint(self, pid: PID_TYPE, tag: Optional[str] = None) -> None:
         """
         Delete a persisted process checkpoint. No error will be raised if
         the checkpoint does not exist
@@ -284,7 +292,7 @@ class PicklePersister(Persister):
         except OSError:
             pass
 
-    def delete_process_checkpoints(self, pid):
+    def delete_process_checkpoints(self, pid: PID_TYPE) -> None:
         """
         Delete all persisted checkpoints related to the given process id
 
@@ -297,24 +305,24 @@ class PicklePersister(Persister):
 class InMemoryPersister(Persister):
     """ Mainly to be used in testing/debugging """
 
-    def __init__(self, loader=None):
+    def __init__(self, loader: Optional[loaders.ObjectLoader] = None) -> None:
         super().__init__()
-        self._checkpoints = {}
+        self._checkpoints: Dict[PID_TYPE, Dict[Optional[str], Bundle]] = {}
         self._save_context = LoadSaveContext(loader=loader)
 
-    def save_checkpoint(self, process, tag=None):
-        self._checkpoints.setdefault(process.pid, {})[tag] = Bundle(process, self._save_context)
+    def save_checkpoint(self, process: 'Process', tag: Optional[str] = None) -> None:
+        self._checkpoints.setdefault(process.pid, {})[tag] = Bundle(process, self._save_context, dereference=True)
 
-    def load_checkpoint(self, pid, tag=None):
+    def load_checkpoint(self, pid: PID_TYPE, tag: Optional[str] = None) -> Bundle:
         return self._checkpoints[pid][tag]
 
-    def get_checkpoints(self):
+    def get_checkpoints(self) -> List[PersistedCheckpoint]:
         cps = []
         for pid in self._checkpoints:
             cps.extend(self.get_process_checkpoints(pid))
         return cps
 
-    def get_process_checkpoints(self, pid):
+    def get_process_checkpoints(self, pid: PID_TYPE) -> List[PersistedCheckpoint]:
         cps = []
         try:
             for tag, _ in self._checkpoints[pid].items():
@@ -323,20 +331,20 @@ class InMemoryPersister(Persister):
             pass
         return cps
 
-    def delete_checkpoint(self, pid, tag=None):
+    def delete_checkpoint(self, pid: PID_TYPE, tag: Optional[str] = None) -> None:
         try:
             del self._checkpoints[pid][tag]
         except KeyError:
             pass
 
-    def delete_process_checkpoints(self, pid):
+    def delete_process_checkpoints(self, pid: PID_TYPE) -> None:
         if pid in self._checkpoints:
             del self._checkpoints[pid]
 
 
-def auto_persist(*members):
+def auto_persist(*members: str) -> Callable[[Type['Savable']], Type['Savable']]:
 
-    def wrapped(savable):
+    def wrapped(savable: Type['Savable']) -> Type['Savable']:
         # pylint: disable=protected-access
         if savable._auto_persist is None:
             savable._auto_persist = set()
@@ -348,7 +356,7 @@ def auto_persist(*members):
     return wrapped
 
 
-def _ensure_object_loader(context, saved_state):
+def _ensure_object_loader(context: Optional['LoadSaveContext'], saved_state: SAVED_STATE_TYPE) -> 'LoadSaveContext':
     """
     Given a LoadSaveContext this method will ensure that it has a valid class loader
     using the following priorities:
@@ -382,23 +390,23 @@ def _ensure_object_loader(context, saved_state):
 
 class LoadSaveContext:
 
-    def __init__(self, loader=None, **kwargs):
+    def __init__(self, loader: Optional[loaders.ObjectLoader] = None, **kwargs: Any) -> None:
         self._values = dict(**kwargs)
         self.loader = loader
 
-    def __getattr__(self, item):
+    def __getattr__(self, item: str) -> Any:
         try:
             return self._values[item]
         except KeyError:
             raise AttributeError("item '{}' not found".format(item))
 
-    def __iter__(self):
+    def __iter__(self) -> Iterable[Any]:
         return self._value.__iter__()
 
-    def __contains__(self, item):
+    def __contains__(self, item: Any) -> bool:
         return self._values.__contains__(item)
 
-    def copyextend(self, **kwargs):
+    def copyextend(self, **kwargs: Any) -> 'LoadSaveContext':
         """ Add additional information to the context by making a copy with the new values """
         extended = self._values.copy()
         extended.update(kwargs)
@@ -406,23 +414,23 @@ class LoadSaveContext:
         return LoadSaveContext(loader=loader, **extended)
 
 
-META = '!!meta'
-META__CLASS_NAME = 'class_name'
-META__OBJECT_LOADER = 'object_loader'
-META__USER = 'user'
-META__TYPES = 'types'
-META__TYPE__METHOD = 'm'
-META__TYPE__SAVABLE = 'S'
+META: str = '!!meta'
+META__CLASS_NAME: str = 'class_name'
+META__OBJECT_LOADER: str = 'object_loader'
+META__USER: str = 'user'
+META__TYPES: str = 'types'
+META__TYPE__METHOD: str = 'm'
+META__TYPE__SAVABLE: str = 'S'
 
 
 class Savable:
-    CLASS_NAME = 'class_name'
+    CLASS_NAME: str = 'class_name'
 
-    _auto_persist = None
+    _auto_persist: Optional[Set[str]] = None
     _persist_configured = False
 
     @staticmethod
-    def load(saved_state, load_context=None):
+    def load(saved_state: SAVED_STATE_TYPE, load_context: Optional[LoadSaveContext] = None) -> 'Savable':
         """
         Load a `Savable` from a saved instance state.  The load context is a way of passing
         runtime data to the object being loaded.
@@ -431,9 +439,10 @@ class Savable:
         :param load_context: Additional runtime state that can be passed into when loading.
             The type and content (if any) is completely user defined
         :return: The loaded Savable instance
-        :rtype: :class:`Savable`
+
         """
         load_context = _ensure_object_loader(load_context, saved_state)
+        assert load_context.loader is not None  # required for type checking
         try:
             class_name = Savable._get_class_name(saved_state)
             load_cls = load_context.loader.load_object(class_name)
@@ -443,45 +452,45 @@ class Savable:
             return load_cls.recreate_from(saved_state, load_context)
 
     @classmethod
-    def auto_persist(cls, *members):
+    def auto_persist(cls, *members: str) -> None:
         if cls._auto_persist is None:
             cls._auto_persist = set()
         cls._auto_persist.update(members)
 
     @classmethod
-    def persist(cls):
+    def persist(cls) -> None:
         pass
 
     @classmethod
-    def recreate_from(cls, saved_state, load_context=None):
+    def recreate_from(cls, saved_state: SAVED_STATE_TYPE, load_context: Optional[LoadSaveContext] = None) -> 'Savable':
         """
         Recreate a :class:`Savable` from a saved state using an optional load context.
 
         :param saved_state: The saved state
         :param load_context: An optional load context
-        :type load_context: :class:`LoadSaveContext`
+
         :return: The recreated instance
-        :rtype: :class:`Savable`
+
         """
         load_context = _ensure_object_loader(load_context, saved_state)
         obj = cls.__new__(cls)
-        base.call_with_super_check(obj.load_instance_state, saved_state, load_context)
+        call_with_super_check(obj.load_instance_state, saved_state, load_context)
         return obj
 
     @super_check
-    def load_instance_state(self, saved_state, load_context):
+    def load_instance_state(self, saved_state: SAVED_STATE_TYPE, load_context: Optional[LoadSaveContext]) -> None:
         self._ensure_persist_configured()
         if self._auto_persist is not None:
             self.load_members(self._auto_persist, saved_state, load_context)
 
     @super_check
-    def save_instance_state(self, out_state, save_context):  # pylint: disable=unused-argument
+    def save_instance_state(self, out_state: SAVED_STATE_TYPE, save_context: Optional[LoadSaveContext]) -> None:  # pylint: disable=unused-argument
         self._ensure_persist_configured()
         if self._auto_persist is not None:
             self.save_members(self._auto_persist, out_state)
 
-    def save(self, save_context=None):
-        out_state = {}
+    def save(self, save_context: Optional[LoadSaveContext] = None) -> SAVED_STATE_TYPE:
+        out_state: SAVED_STATE_TYPE = {}
 
         if save_context is None:
             save_context = LoadSaveContext()
@@ -498,10 +507,10 @@ class Savable:
             loader = default_loader
 
         Savable._set_class_name(out_state, loader.identify_object(self.__class__))
-        base.call_with_super_check(self.save_instance_state, out_state, save_context)
+        call_with_super_check(self.save_instance_state, out_state, save_context)
         return out_state
 
-    def save_members(self, members, out_state):
+    def save_members(self, members: Iterable[str], out_state: SAVED_STATE_TYPE) -> None:
         for member in members:
             value = getattr(self, member)
             if inspect.ismethod(value):
@@ -516,11 +525,16 @@ class Savable:
                 value = copy.deepcopy(value)
             out_state[member] = value
 
-    def load_members(self, members, saved_state, load_context=None):
+    def load_members(
+        self,
+        members: Iterable[str],
+        saved_state: SAVED_STATE_TYPE,
+        load_context: Optional[LoadSaveContext] = None
+    ) -> None:
         for member in members:
             setattr(self, member, self._get_value(saved_state, member, load_context))
 
-    def _ensure_persist_configured(self):
+    def _ensure_persist_configured(self) -> None:
         if not self._persist_configured:
             self.persist()
             self._persist_configured = True
@@ -528,36 +542,36 @@ class Savable:
     # region Metadata getter/setters
 
     @staticmethod
-    def set_custom_meta(out_state, name, value):
+    def set_custom_meta(out_state: SAVED_STATE_TYPE, name: str, value: Any) -> None:
         user_dict = Savable._get_create_meta(out_state).setdefault(META__USER, {})
         user_dict[name] = value
 
     @staticmethod
-    def get_custom_meta(saved_state, name):
+    def get_custom_meta(saved_state: SAVED_STATE_TYPE, name: str) -> Any:
         try:
             return saved_state[META][name]
         except KeyError:
             raise ValueError("Unknown meta key '{}'".format(name))
 
     @staticmethod
-    def _get_create_meta(out_state):
+    def _get_create_meta(out_state: SAVED_STATE_TYPE) -> Dict[str, Any]:
         return out_state.setdefault(META, {})
 
     @staticmethod
-    def _set_class_name(out_state, name):
+    def _set_class_name(out_state: SAVED_STATE_TYPE, name: str) -> None:
         Savable._get_create_meta(out_state)[META__CLASS_NAME] = name
 
     @staticmethod
-    def _get_class_name(saved_state):
+    def _get_class_name(saved_state: SAVED_STATE_TYPE) -> str:
         return Savable._get_create_meta(saved_state)[META__CLASS_NAME]
 
     @staticmethod
-    def _set_meta_type(out_state, name, type_spec):
+    def _set_meta_type(out_state: SAVED_STATE_TYPE, name: str, type_spec: Any) -> None:
         type_dict = Savable._get_create_meta(out_state).setdefault(META__TYPES, {})
         type_dict[name] = type_spec
 
     @staticmethod
-    def _get_meta_type(saved_state, name):
+    def _get_meta_type(saved_state: SAVED_STATE_TYPE, name: str) -> Any:
         try:
             return saved_state[META][META__TYPES][name]
         except KeyError:
@@ -565,7 +579,8 @@ class Savable:
 
     # endregion
 
-    def _get_value(self, saved_state, name, load_context):
+    def _get_value(self, saved_state: SAVED_STATE_TYPE, name: str,
+                   load_context: Optional[LoadSaveContext]) -> Union[MethodType, 'Savable']:
         value = saved_state[name]
 
         typ = Savable._get_meta_type(saved_state, name)
@@ -585,21 +600,21 @@ class SavableFuture(futures.Future, Savable):
     .. note: This does not save any assigned done callbacks.
     """
 
-    def save_instance_state(self, out_state, save_context):
+    def save_instance_state(self, out_state: SAVED_STATE_TYPE, save_context: LoadSaveContext) -> None:
         super().save_instance_state(out_state, save_context)
         if self.done() and self.exception() is not None:
             out_state['exception'] = self.exception()
 
     @classmethod
-    def recreate_from(cls, saved_state, load_context=None):
+    def recreate_from(cls, saved_state: SAVED_STATE_TYPE, load_context: Optional[LoadSaveContext] = None) -> 'Savable':
         """
         Recreate a :class:`Savable` from a saved state using an optional load context.
 
         :param saved_state: The saved state
         :param load_context: An optional load context
-        :type load_context: :class:`LoadSaveContext`
+
         :return: The recreated instance
-        :rtype: :class:`Savable`
+
         """
         load_context = _ensure_object_loader(load_context, saved_state)
 
@@ -610,10 +625,10 @@ class SavableFuture(futures.Future, Savable):
 
         state = saved_state['_state']
 
-        if state == asyncio.futures._PENDING:  # pylint: disable=protected-access
+        if state == asyncio.futures._PENDING:  # type: ignore # pylint: disable=protected-access
             obj = cls(loop=loop)
 
-        if state == asyncio.futures._FINISHED:  # pylint: disable=protected-access
+        if state == asyncio.futures._FINISHED:  # type: ignore # pylint: disable=protected-access
             obj = cls(loop=loop)
             result = saved_state['_result']
 
@@ -623,16 +638,16 @@ class SavableFuture(futures.Future, Savable):
             except KeyError:
                 obj.set_result(result)
 
-        if state == asyncio.futures._CANCELLED:  # pylint: disable=protected-access
+        if state == asyncio.futures._CANCELLED:  # type: ignore # pylint: disable=protected-access
             obj = cls(loop=loop)
             obj.cancel()
 
         return obj
 
-    def load_instance_state(self, saved_state, load_context):
+    def load_instance_state(self, saved_state: SAVED_STATE_TYPE, load_context: LoadSaveContext) -> None:
         # pylint: disable=attribute-defined-outside-init
         super().load_instance_state(saved_state, load_context)
-
         if self._callbacks:
-            for callback in self._callbacks:
+            # typing says asyncio.Future._callbacks needs to be called, but in the python 3.7 code it is a simple list
+            for callback in self._callbacks:  # type: ignore
                 self.remove_done_callback(callback)
